/*
 * This file is part of the Extra-P software (http://www.scalasca.org/software/extra-p)
 *
 * Copyright (c) 2016-2018,
 * Technische Universitaet Darmstadt, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 *
 */

#include <config.h>
#include <EXTRAP_Interface.hpp>
#include <EXTRAP_IncrementalPoint.hpp>
#include <EXTRAP_Model.hpp>
#include <EXTRAP_SingleParameterSimpleModelGenerator.hpp>
#include <EXTRAP_SingleParameterRefiningModelGenerator.hpp>
#include <EXTRAP_MultiParameterSimpleModelGenerator.hpp>
#include <EXTRAP_Utilities.hpp>
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <string>
#include <sstream>
#include <stdlib.h>
#include <fstream>
#include <iostream>
#include <algorithm>

namespace EXTRAP
{
Experiment::Experiment( void )
{
}

Experiment::Experiment( const Experiment& e )
{
}

Experiment&
Experiment::operator=( const Experiment& e )
{
    return *this;
}

Experiment::~Experiment( void )
{
    // Delete metric definitions
    for ( MetricList::iterator i = m_metrics.begin();
          i != m_metrics.end();
          i++ )
    {
        delete( *i );
    }

    // Delete region definitions
    for ( RegionList::iterator i = m_regions.begin();
          i != m_regions.end();
          i++ )
    {
        delete( *i );
    }

    // Delete callpath definitions
    for ( CallpathList::iterator i = m_all_callpaths.begin();
          i != m_all_callpaths.end();
          i++ )
    {
        delete( *i );
    }

    // Delete coordinate collection
    for ( CoordinateList::iterator i = m_coordinates.begin();
          i != m_coordinates.end();
          i++ )
    {
        delete( *i );
    }

    // Delete all associated models
    for ( ModelMatrix::iterator i = m_models.begin();
          i != m_models.end();
          i++ )
    {
        for ( std::vector<ModelList>::iterator j = ( *i ).begin();
              j != ( *i ).end();
              j++ )
        {
            for ( ModelList::iterator k = ( *j ).begin();
                  k != ( *j ).end();
                  k++ )
            {
                delete( *k );
            }
        }
    }

    // Delete all data points
    for ( ExperimentPointMatrix::iterator i = m_points.begin();
          i != m_points.end();
          i++ )
    {
        for ( std::vector<ExperimentPointList>::iterator j = ( *i ).begin();
              j != ( *i ).end();
              j++ )
        {
            for ( ExperimentPointList::iterator k = ( *j ).begin();
                  k != ( *j ).end();
                  k++ )
            {
                delete( *k );
            }
        }
    }

    // Delete all model comments
    for ( ModelCommentList::iterator i = m_comments.begin();
          i != m_comments.end();
          i++ )
    {
        delete( *i );
    }

    // Delete all model generators
    for ( ModelGeneratorList::iterator i = m_generators.begin();
          i != m_generators.end();
          i++ )
    {
        delete( *i );
    }
}

const MetricList&
Experiment::getMetrics( void ) const
{
    return m_metrics;
}

const RegionList&
Experiment::getRegions( void ) const
{
    return m_regions;
}

const CallpathList&
Experiment::getRootCallpaths( void ) const
{
    return m_root_callpaths;
}

const CallpathList&
Experiment::getAllCallpaths( void ) const
{
    return m_all_callpaths;
}

const ParameterList&
Experiment::getParameters( void ) const
{
    return m_parameters;
}

const ModelList&
Experiment::getModels( const Metric&   metric,
                       const Callpath& callpath ) const
{
    // Use at() because it performs boundary checking.
    // The [] operator does not.
    if ( m_models.size() > metric.getId() &&
         m_models.at( metric.getId() ).size() > callpath.getId() )
    {
        return m_models.at( metric.getId() ).at( callpath.getId() );
    }
    return m_empty_model_list;
}

const ExperimentPointList&
Experiment::getPoints( const Metric&   metric,
                       const Callpath& callpath ) const
{
    // Use at() because it performs boundary checking.
    // The [] operator does not.
    return m_points.at( metric.getId() ).at( callpath.getId() );
}

const ModelGeneratorList&
Experiment::getModelGenerators( void ) const
{
    return m_generators;
}

const ModelCommentList&
Experiment::getModelComments( void ) const
{
    return m_comments;
}

Parameter
Experiment::getParameter( int id ) const
{
    assert( id >= 0 && id < m_parameters.size() );
    return m_parameters[ id ];
}

Metric*
Experiment::getMetric( int id ) const
{
    assert( id >= 0 && id < m_metrics.size() );
    return m_metrics[ id ];
}

Region*
Experiment::getRegion( int id ) const
{
    assert( id >= 0 && id < m_regions.size() );
    return m_regions[ id ];
}

Callpath*
Experiment::getCallpath( int id ) const
{
    assert( id >= 0 && id < m_all_callpaths.size() );
    return m_all_callpaths[ id ];
}

const Coordinate*
Experiment::getCoordinate( int id ) const
{
    assert( id >= 0 && id < m_coordinates.size() );
    return m_coordinates[ id ];
}

ModelGenerator*
Experiment::getModelGenerator( int id ) const
{
    assert( id >= 0 && id < m_generators.size() );
    return m_generators[ id ];
}

ModelComment*
Experiment::getModelComment( int id ) const
{
    assert( id >= 0 && id < m_comments.size() );
    return m_comments[ id ];
}

void
Experiment::addParameter( Parameter& newParameter )
{
    newParameter.setId( m_parameters.size() );
    m_parameters.push_back( newParameter );
}

void
Experiment::addMetric( Metric* newMetric )
{
    if ( newMetric == NULL )
    {
        return;
    }
    newMetric->setId( m_metrics.size() );
    m_metrics.push_back( newMetric );
}

void
Experiment::addRegion( Region* newRegion )
{
    if ( newRegion == NULL )
    {
        return;
    }
    newRegion->setId( m_regions.size() );
    m_regions.push_back( newRegion );
}

void
Experiment::addCoordinate( Coordinate* newCoordinate )
{
    if ( newCoordinate == NULL )
    {
        return;
    }
    newCoordinate->setId( m_coordinates.size() );
    m_coordinates.push_back( newCoordinate );
}

void
Experiment::addCallpath( Callpath* newCallpath )
{
    if ( newCallpath == NULL )
    {
        return;
    }
    newCallpath->setId( m_all_callpaths.size() );
    m_all_callpaths.push_back( newCallpath );
    if ( newCallpath->getParent() == NULL )
    {
        m_root_callpaths.push_back( newCallpath );
    }
}

void
Experiment::addDataPoint( ExperimentPoint* newDataPoint,
                          const Metric&    metric,
                          const Callpath&  callpath )
{
    if ( newDataPoint == NULL )
    {
        return;
    }
    for ( int i = m_points.size(); i <= metric.getId(); i++ )
    {
        m_points.push_back( std::vector<ExperimentPointList>() );
    }
    for ( int i = m_points[ metric.getId() ].size(); i <= callpath.getId(); i++ )
    {
        m_points[ metric.getId() ].push_back( ExperimentPointList() );
    }
    m_points[ metric.getId() ][ callpath.getId() ].push_back( newDataPoint );
}

void
Experiment::addModel( Model*          newModel,
                      const Metric&   metric,
                      const Callpath& callpath )
{
    if ( newModel == NULL )
    {
        return;
    }
    for ( int i = m_models.size(); i <= metric.getId(); i++ )
    {
        m_models.push_back( std::vector<ModelList>() );
    }
    for ( int i = m_models[ metric.getId() ].size(); i <= callpath.getId(); i++ )
    {
        m_models[ metric.getId() ].push_back( ModelList() );
    }
    ModelCommentList commentList = newModel->getComments();
    ModelCommentList tempList;
    for ( int i = 0; i < commentList.size(); i++ )
    {
        tempList.push_back( this->addModelComment( commentList[ i ]->getMessage() ) );
    }
    newModel->clearComments();
    for ( int i = 0; i < tempList.size(); i++ )
    {
        newModel->addComment( tempList[ i ] );
    }

    m_models[ metric.getId() ][ callpath.getId() ].push_back( newModel );
}

void
Experiment::addModelGenerator( ModelGenerator* newGenerator )
{
    if ( newGenerator == NULL )
    {
        return;
    }
    newGenerator->setId( m_generators.size() );
    m_generators.push_back( newGenerator );
}

void
Experiment::addModelComment( ModelComment* newComment )
{
    if ( newComment == NULL )
    {
        return;
    }
    newComment->setId( m_comments.size() );
    m_comments.push_back( newComment );
}

ModelComment*
Experiment::addModelComment( const std::string& message )
{
    for ( ModelCommentList::iterator it = m_comments.begin();
          it != m_comments.end();
          it++ )
    {
        if ( ( *it )->getMessage() == message )
        {
            return *it;
        }
    }
    ModelComment* newComment = new ModelComment( message );
    addModelComment( newComment );
    return newComment;
}

void
Experiment::modelAll( ModelGenerator& generator )
{
    for ( MetricList::iterator metric_it = m_metrics.begin();
          metric_it != m_metrics.end();
          metric_it++ )
    {
        Metric* metric = *metric_it;
        for ( CallpathList::iterator callpath_it = m_all_callpaths.begin();
              callpath_it != m_all_callpaths.end();
              callpath_it++ )
        {
            Callpath*                  callpath    = *callpath_it;
            const ExperimentPointList& data_points = getPoints( *metric, *callpath );
            Model*                     model       = generator.createModel( data_points );
            assert( model->getModelFunction() );
            addModel( model, *metric, *callpath );
        }
    }
}

void
Experiment::deleteModel( int             modelIndex,
                         const Metric&   metric,
                         const Callpath& callpath )
{
    ModelList& models = ( ModelList& )getModels( metric, callpath );
    Model*     model  = models[ modelIndex ];
    models.erase( models.begin() + modelIndex );
    delete( model );
}

void
Experiment::deleteModel( int modelIndex )
{
    for ( MetricList::iterator metric_it = m_metrics.begin();
          metric_it != m_metrics.end();
          metric_it++ )
    {
        Metric* metric = *metric_it;
        for ( CallpathList::iterator callpath_it = m_all_callpaths.begin();
              callpath_it != m_all_callpaths.end();
              callpath_it++ )
        {
            deleteModel( modelIndex,
                         *metric,
                         **callpath_it );
        }
    }
}

bool
Experiment::writeExtrapFile( const std::string& filename ) const
{
    std::ofstream dataFile( filename.c_str(), std::ios::binary );
    if ( !serialize( dataFile ) )
    {
        ErrorStream << "Could not write File successfully" << std::endl;
        dataFile.close();
        remove( filename.c_str() );
        return false;
    }
    else
    {
        dataFile.close();
        return true;
    }
}

Experiment*
Experiment::openExtrapFile( const std::string& filename )
{
    std::ifstream dataFile( filename.c_str(), std::ios::binary );
    Experiment*   new_experiment = deserialize( dataFile );
    if ( new_experiment == NULL )
    {
        ErrorStream << "Could not read File successfully" << std::endl;
    }
    dataFile.close();
    return new_experiment;
}

Experiment*
Experiment::openTextInput( const std::string& filename )
{
    std::ifstream dataFile( filename.c_str() );
    std::string   lineString;
    std::string   fieldName;
    //  int state=0;
    if ( dataFile.is_open() == false )
    {
        ErrorStream << "Unable to open the test data file " << filename << std::endl;

        return NULL;
    }
    Experiment*      new_experiment = new Experiment();
    Metric*          new_metric     = NULL;
    Parameter        parameter;
    CoordinateList   pointslist;
    Region*          new_region = NULL;
    Region*          old_region = NULL;
    Callpath*        callpath;
    IncrementalPoint data_point;
    int              iterator = 0;

    int currentLine = 0;
    while ( !dataFile.eof() )
    {
        // get a line and store in lineString
        getline( dataFile, lineString, '\n' );
        currentLine++;

        // ignore empty lines
        if ( lineString.empty() )
        {
            continue;
        }
        // ignore comments
        if ( lineString.c_str()[ 0 ] == '#' )
        {
            continue;
        }

        // get name of field
        std::istringstream iss( lineString );
        iss >> fieldName;
        std::string tmp;

        if ( fieldName == "METRIC" )
        {
            tmp.clear();
            iss >> tmp;
            new_metric = new Metric( tmp, "unknown" );
            new_experiment->addMetric( new_metric );
        }
        else if ( fieldName == "REGION" )
        {
            tmp.clear();
            iss >> tmp;
            new_region = new Region( tmp, filename, 1 );
            new_experiment->addRegion( new_region );
            callpath = new Callpath( new_region, NULL );
            new_experiment->addCallpath( callpath );
        }
        else if ( fieldName == "DATA" )
        {
            if ( new_metric == NULL )
            {
                new_metric = new Metric( "Test", "unknown" );
                new_experiment->addMetric( new_metric );
            }
            data_point.clear();
            //double test;
            //iss >> test;
            while ( !iss.eof() )
            {
                double measured_point;
                iss >> measured_point;
                data_point.addValue( measured_point );
            }
            if ( new_region == NULL )
            {
                ErrorStream << "In line "
                            << currentLine
                            << ": experiment line missing "
                            << std::endl;

                return NULL;
            }
            if ( ( iterator == 0 ) && ( old_region == new_region ) && ( old_region != NULL ) )
            {
                ErrorStream << "In line "
                            << currentLine
                            << ": too many data points lines before new experiment  "
                            << std::endl;

                return NULL;
            }
            if ( ( iterator != 0 ) && ( old_region != new_region ) && ( old_region != NULL ) )
            {
                ErrorStream << "In line "
                            << currentLine
                            << ": too few data points lines before new experiment "
                            << std::endl;

                return NULL;
            }
            if ( iterator >= pointslist.size() )
            {
                ErrorStream << "In line "
                            << currentLine
                            << ": measurement points should be defined "
                            << std::endl;

                return NULL;
            }
            new_experiment->addDataPoint( data_point.getExperimentPoint( pointslist[ iterator ],
                                                                         new_metric,
                                                                         callpath ),
                                          *new_metric,
                                          *callpath );
            iterator++;
            if ( iterator == pointslist.size() )
            {
                iterator = 0;
            }
            old_region = new_region;
        }

        else if ( fieldName == "PARAMETER" )
        {
            tmp.clear();
            iss >> tmp;
            parameter = Parameter( tmp );
            new_experiment->addParameter( parameter );
        }

        else if ( fieldName == "POINTS" )
        {
            double old_value;
            double value;
            bool   first = true;
            while ( !iss.eof() )
            {
                iss >> value;
                if ( first == true )
                {
                    first     = false;
                    old_value = value;
                    Coordinate* pv_list = new Coordinate();
                    ( *pv_list )[ parameter ] = value;
                    pointslist.push_back( pv_list );
                    new_experiment->addCoordinate( pv_list );
                }
                else if ( old_value != value )
                {
                    Coordinate* pv_list = new Coordinate();
                    ( *pv_list )[ parameter ] = value;
                    pointslist.push_back( pv_list );
                    new_experiment->addCoordinate( pv_list );
                    old_value = value;
                }
            }
        }
        else
        {
            ErrorStream << fieldName
                        << " not a valid test file row start value.";

            return NULL;
        }
    }

    dataFile.close();

    SingleParameterSimpleModelGenerator* generator = new SingleParameterSimpleModelGenerator();
    MetricList                           metrics   = new_experiment->getMetrics();
    CallpathList                         callpaths = new_experiment->getAllCallpaths();
    generator->setEpsilon( 0.05 );
    generator->setMaxTermCount( 1 );
    generator->generateDefaultHypothesisBuildingBlockSet();
    new_experiment->addModelGenerator( generator );

    for ( MetricList::iterator metric_it = metrics.begin();
          metric_it != metrics.end();
          metric_it++ )
    {
        Metric* metric = *metric_it;
        int     c      = 0;
        for ( CallpathList::iterator callpath_it = callpaths.begin();
              callpath_it != callpaths.end();
              callpath_it++ )
        {
            Callpath* callpath = *callpath_it;

            const ExperimentPointList data_points = new_experiment->getPoints( *metric, *callpath );
            Model*                    model       = generator->createModel( data_points );
            new_experiment->addModel( model, *metric, *callpath );
            c++;
        }
    }

    return new_experiment;
}

bool
Experiment::serialize( std::ofstream& stream ) const
{
    IoHelper ioHelper = IoHelper( &stream );
    //Write Prefix and Version Specifier
    SAFE_RETURN( ioHelper.writeString( EXTRAP_CLASSIFIER ) );
    SAFE_RETURN( ioHelper.writeString( VERSION_QUALIFIER ) );
    //Write Parameters
    for ( ParameterList::const_iterator it = this->m_parameters.begin(); it != this->m_parameters.end(); it++ )
    {
        Parameter p = *it;
        SAFE_RETURN(  p.serialize(  &ioHelper ) );
    }
    for ( MetricList::const_iterator it = this->m_metrics.begin(); it != this->m_metrics.end(); it++ )
    {
        Metric* m = *it;
        SAFE_RETURN(  m->serialize(  &ioHelper ) );
    }
    for ( RegionList::const_iterator it = this->m_regions.begin(); it != this->m_regions.end(); it++ )
    {
        Region* r = *it;
        SAFE_RETURN(  r->serialize(  &ioHelper ) );
    }
    for ( CallpathList::const_iterator it = this->m_all_callpaths.begin(); it != this->m_all_callpaths.end(); it++ )
    {
        Callpath* cp = *it;
        SAFE_RETURN(  cp->serialize(  &ioHelper ) );
    }
    for ( CoordinateList::const_iterator it = this->m_coordinates.begin(); it != this->m_coordinates.end(); it++ )
    {
        SAFE_RETURN(  ( *it )->serialize(  &ioHelper ) );
    }
    for ( ModelCommentList::const_iterator it = this->m_comments.begin(); it != this->m_comments.end(); it++ )
    {
        SAFE_RETURN(  ( *it )->serialize(  &ioHelper ) );
    }
    for ( ModelGeneratorList::const_iterator it = this->m_generators.begin(); it != this->m_generators.end(); it++ )
    {
        SAFE_RETURN(  ( *it )->serialize(  &ioHelper ) );
    }
    for ( MetricList::const_iterator it = this->m_metrics.begin(); it != this->m_metrics.end(); it++ )
    {
        for ( CallpathList::const_iterator it2 = this->m_all_callpaths.begin(); it2 != this->m_all_callpaths.end(); it2++ )
        {
            ModelList list = this->getModels( **it, **it2 );
            for ( ModelList::iterator it3 = list.begin(); it3 != list.end(); it3++ )
            {
                SAFE_RETURN(  ( *it3 )->serialize( *it, *it2, &ioHelper ) );
            }
        }
    }
    int points = 0;
    for ( MetricList::const_iterator it = this->m_metrics.begin(); it != this->m_metrics.end(); it++ )
    {
        for ( CallpathList::const_iterator it2 = this->m_all_callpaths.begin(); it2 != this->m_all_callpaths.end(); it2++ )
        {
            ExperimentPointList list = this->getPoints( **it, **it2 );
            points += list.size();
            for ( ExperimentPointList::const_iterator it3 = list.begin(); it3 != list.end(); it3++ )
            {
                SAFE_RETURN(  ( *it3 )->serialize(  &ioHelper ) );
            }
        }
    }
    return true;
}

Experiment*
Experiment::deserialize( std::ifstream& stream )
{
    IoHelper    ioHelper = IoHelper( &stream );
    Experiment* exp      = new Experiment();
    //Write Prefix and Version Specifier
    std::string qualifier = ioHelper.readString();
    if ( 0 != qualifier.compare( EXTRAP_CLASSIFIER ) )
    {
        ErrorStream << "This is not an EXTRAP Experiment File." << std::endl;
        return NULL;
    }
    std::string versionNumber = ioHelper.readString();
    char        c;
    std::string prefix = ioHelper.readString();
    while ( stream.good() )
    {
        DebugStream << "Deserialize " << prefix << std::endl;
        if ( 0 == prefix.compare( Parameter::PARAMETER_PREFIX ) )
        {
            Parameter p = Parameter::deserialize( &ioHelper );
            exp->addParameter( p );
        }
        else if ( 0 == prefix.compare( Metric::METRIC_PREFIX ) )
        {
            Metric* m = Metric::deserialize( &ioHelper );
            SAFE_RETURN_NULL(  m );
            exp->addMetric( m );
        }
        else if ( 0 == prefix.compare( Region::REGION_PREFIX ) )
        {
            Region* r = Region::deserialize( &ioHelper );
            SAFE_RETURN_NULL(  r );
            exp->addRegion( r );
        }
        else if ( 0 == prefix.compare( Callpath::CALLPATH_PREFIX ) )
        {
            Callpath* cp = Callpath::deserialize( exp, &ioHelper );
            exp->addCallpath( cp );
        }
        else if ( 0 == prefix.compare( Coordinate::COORDINATE_PREFIX ) )
        {
            Coordinate* c = Coordinate::deserialize( exp, &ioHelper );
            SAFE_RETURN_NULL(  c );
            exp->addCoordinate( c );
        }
        else if ( 0 == prefix.compare( ModelComment::MODELCOMMENT_PREFIX ) )
        {
            ModelComment* comment = ModelComment::deserialize( &ioHelper );
            SAFE_RETURN_NULL(  comment );
            exp->addModelComment( comment );
        }
        else if ( 0 == prefix.compare( SingleParameterSimpleModelGenerator::SINGLEPARAMETERSIMPLEMODELGENERATOR_PREFIX ) )
        {
            ModelGenerator* generator = SingleParameterSimpleModelGenerator::deserialize( &ioHelper );
            SAFE_RETURN_NULL(  generator );
            exp->addModelGenerator( generator );
        }
        else if ( 0 == prefix.compare( SingleParameterRefiningModelGenerator::SINGLEPARAMETERREFININGMODELGENERATOR_PREFIX ) )
        {
            ModelGenerator* generator = SingleParameterRefiningModelGenerator::deserialize( &ioHelper );
            SAFE_RETURN_NULL(  generator );
            exp->addModelGenerator( generator );
        }
        else if ( 0 == prefix.compare( MultiParameterSimpleModelGenerator::MULTIPARAMETERSIMPLEMODELGENERATOR_PREFIX ) )
        {
            ModelGenerator* generator = MultiParameterSimpleModelGenerator::deserialize( &ioHelper );
            SAFE_RETURN_NULL(  generator );
            exp->addModelGenerator( generator );
        }
        else if ( 0 == prefix.compare( ExperimentPoint::EXPERIMENTPOINT_PREFIX ) )
        {
            ExperimentPoint* point = ExperimentPoint::deserialize( exp, &ioHelper );
            SAFE_RETURN_NULL(  point );
            exp->addDataPoint( point,
                               *point->getMetric(),
                               *point->getCallpath() );
        }
        else if ( 0 == prefix.compare( Model::MODEL_PREFIX ) )
        {
            int64_t metric_id;
            int64_t callpath_id;
            Model*  model = Model::deserialize( exp, &metric_id, &callpath_id, &ioHelper );
            SAFE_RETURN_NULL(  model );
            exp->addModel( model,
                           *exp->getMetric( metric_id ),
                           *exp->getCallpath( callpath_id ) );
        }
        else
        {
            ErrorStream << "Unknown object: " << prefix << ". Can not load experiment." << std::endl;
            return NULL;
        }

        prefix = ioHelper.readString();
    }
    return exp;
}

bool
equal( const Experiment* lhs, const Experiment* rhs )
{
    if ( lhs == rhs )
    {
        return true;
    }
    else if ( lhs == NULL || rhs == NULL )
    {
        return false;
    }
    bool                         result        = true;
    const std::vector<Parameter> parametersLhs = lhs->getParameters();
    const std::vector<Parameter> parametersRhs = rhs->getParameters();
    if ( parametersLhs.size() != parametersRhs.size() )
    {
        return false;
    }
    int length = parametersLhs.size();
    for ( int i = 0; i < length; i++ )
    {
        if ( !equal( &parametersLhs[ i ], &parametersRhs[ i ] ) )
        {
            return false;
        }
    }
    const std::vector<Region*> regionsLhs = lhs->getRegions();
    const std::vector<Region*> regionsRhs = rhs->getRegions();
    if ( regionsLhs.size() != regionsRhs.size() )
    {
        return false;
    }
    int regionsLength = regionsLhs.size();
    for ( int i = 0; i < regionsLength; i++ )
    {
        if ( !equal( regionsLhs[ i ], regionsRhs[ i ] ) )
        {
            return false;
        }
    }
    const std::vector<Metric*> metricsLhs = lhs->getMetrics();
    const std::vector<Metric*> metricsRhs = rhs->getMetrics();
    if ( metricsLhs.size() != metricsRhs.size() )
    {
        return false;
    }
    int                          lengthMetrics = metricsLhs.size();
    const std::vector<Callpath*> callpathsLhs  = lhs->getAllCallpaths();
    const std::vector<Callpath*> callpathsRhs  = rhs->getAllCallpaths();
    if ( callpathsLhs.size() != callpathsRhs.size() )
    {
        return false;
    }
    int lengthCallpaths = callpathsLhs.size();
    for ( int i = 0; i < lengthMetrics; i++ )
    {
        if ( !equal( metricsLhs[ i ], metricsRhs[ i ] ) )
        {
            return false;
        }
        for ( int j = 0; j < lengthCallpaths; j++ )
        {
            if ( !equal( callpathsLhs[ j ], callpathsRhs[ j ] ) )
            {
                return false;
            }
            ExperimentPointList pointsLhs = lhs->getPoints( *metricsLhs[ i ], *callpathsLhs[ j ] );
            ExperimentPointList pointsRhs = rhs->getPoints( *metricsRhs[ i ], *callpathsRhs[ j ] );
            if ( pointsLhs.size() != pointsRhs.size() )
            {
                return false;
            }

            int lengthPoints = pointsLhs.size();
            for ( int k = 0; k < lengthPoints; k++ )
            {
                if ( !equal( pointsLhs[ k ], pointsRhs[ k ] ) )
                {
                    return false;
                }
            }

            ModelList modelsLhs = lhs->getModels( *metricsLhs[ i ], *callpathsLhs[ j ] );
            ModelList modelsRhs = rhs->getModels( *metricsRhs[ i ], *callpathsRhs[ j ] );
            if ( modelsLhs.size() != modelsRhs.size() )
            {
                return false;
            }
            int lengthModels = modelsLhs.size();
            for ( int l = 0; l < lengthModels; l++ )
            {
                if ( !equal( modelsLhs[ l ], modelsRhs[ l ] ) )
                {
                    return false;
                }
            }
        }
    }
    return true;
}
};
