/*
 * This file is part of the Extra-P software (http://www.scalasca.org/software/extra-p)
 *
 * Copyright (c) 2017,
 * Technische Universitaet Darmstadt, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 *
 */
#include <config.h>
#include <EXTRAP_MultiParameterModelGenerator.hpp>
#include <EXTRAP_MultiParameterSimpleFunctionModeler.hpp>
#include <EXTRAP_MultiParameterHypothesis.hpp>
#include <EXTRAP_MultiParameterFunction.hpp>
#include <EXTRAP_Interface.hpp>
#include <EXTRAP_Utilities.hpp>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cassert>

namespace EXTRAP
{
MultiParameterModelGenerator::MultiParameterModelGenerator()
{
    m_options = GENERATE_MODEL_MEDIAN;
}

Model*
MultiParameterModelGenerator::createModel( const ExperimentPointList& modeledPointList,
                                           const Function*            expectationFunction )
{
    FunctionPair* confidence_interval = new FunctionPair();
    FunctionPair* error_cone          = new FunctionPair();
    FunctionPair* noise_error         = new FunctionPair();

    confidence_interval->upper = new Function();
    confidence_interval->lower = new Function();
    error_cone->upper          = new Function();
    error_cone->lower          = new Function();
    noise_error->upper         = new Function();
    noise_error->lower         = new Function();

    // Build list of data points from experiment points
    std::vector<DataPoint> dataPoints;
    dataPoints.reserve( modeledPointList.size() );
    for ( std::vector<ExperimentPoint*>::const_iterator it = modeledPointList.begin(); it != modeledPointList.end(); ++it )
    {
        ExperimentPoint* p = *it;
        if ( m_options == GENERATE_MODEL_MEAN )
        {
            dataPoints.push_back( DataPoint( &( p->getCoordinate() ), p->getSampleCount(), p->getMean(), p->getMeanCI() ) );
        }
        else if ( m_options == GENERATE_MODEL_MEDIAN )
        {
            dataPoints.push_back( DataPoint( &( p->getCoordinate() ), p->getSampleCount(), p->getMedian(), p->getMedianCI() ) );
        }
        else
        {
            assert( false );
        }
    }
    //std::cout<<"here-in-generato0r"<<std::endl;

    ModelCommentList         comments       = ModelCommentList();
    MultiParameterHypothesis bestHypothesis = getFunctionModeler().createModel( dataPoints, comments, expectationFunction );
    // std::cout<<"here-in-generator"<<std::endl;
    // The resulting model will take ownership of the function
    return new Model( ( Function* )bestHypothesis.getFunction(),
                      this,
                      confidence_interval,
                      error_cone,
                      noise_error,
                      comments,
                      bestHypothesis.getRSS(),
                      bestHypothesis.getAR2(),
                      bestHypothesis.getSMAPE() );
}

void
MultiParameterModelGenerator::setModelOptions( GenerateModelOptions options )
{
    this->m_options = options;
}

GenerateModelOptions
MultiParameterModelGenerator::getModelOptions( void ) const
{
    return this->m_options;
}

bool
MultiParameterModelGenerator::serialize( IoHelper* ioHelper ) const
{
    SAFE_RETURN( ioHelper->writeString(  m_user_name ) );
    GenerateModelOptions options = this->getModelOptions();
    std::string          optionsString;
    switch ( options )
    {
        case GENERATE_MODEL_MEAN:
            optionsString = "GENERATE_MODEL_MEAN";
            break;
        case GENERATE_MODEL_MEDIAN:
            optionsString = "GENERATE_MODEL_MEDIAN";
            break;
        default:
            ErrorStream << "Invalid GenerateModelOptions found in ModelGenerator. Exiting..." << std::endl;
            return false;
    }
    SAFE_RETURN( ioHelper->writeString(  optionsString ) );
    return true;
}

void
MultiParameterModelGenerator::deserialize( IoHelper* ioHelper )
{
    std::string userName = ioHelper->readString();
    this->setUserName( userName );
    std::string modelOptions = ioHelper->readString();
    if ( modelOptions.compare( "GENERATE_MODEL_MEAN" ) == 0 )
    {
        this->setModelOptions( GENERATE_MODEL_MEAN );
    }
    else if ( modelOptions.compare( "GENERATE_MODEL_MEDIAN" ) == 0 )
    {
        this->setModelOptions( GENERATE_MODEL_MEDIAN );
    }
    else
    {
        ErrorStream << "Invalid ModelOptions found in File. Using Mean" << std::endl;
        this->setModelOptions( GENERATE_MODEL_MEAN );
    }
}
}; // Close namespace
