/*
 * This file is part of the Extra-P software (http://www.scalasca.org/software/extra-p)
 *
 * Copyright (c) 2017,
 * Technische Universitaet Darmstadt, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 *
 */

#include <config.h>
#include <EXTRAP_SingleParameterHypothesis.hpp>
#include <EXTRAP_Utilities.hpp>
#include <cmath>
#include <sstream>
#include <limits>
#include <cassert>

namespace EXTRAP
{
SingleParameterHypothesis::SingleParameterHypothesis( void ) : m_function( NULL ), m_function_refcount( NULL )
{
}

SingleParameterHypothesis::~SingleParameterHypothesis()
{
}

SingleParameterHypothesis::SingleParameterHypothesis( SingleParameterFunction* f )
{
    this->m_function                 = f;
    this->m_function_refcount        = new HypothesisFunctionRefcount();
    this->m_function_refcount->count = 1; // Initially referenced only inside this hypothesis
    this->m_function_refcount->func  = f;
    DebugStream << "### Initialized refcount to 1 for " << this->m_function << " at " << this->m_function_refcount << "\n";
}

SingleParameterFunction*
SingleParameterHypothesis::getFunction( void ) const
{
    return this->m_function;
}

double
SingleParameterHypothesis::getRSS( void ) const
{
    return this->m_RSS;
}

void
SingleParameterHypothesis::setRSS( double rss )
{
    this->m_RSS = rss;
}

double
SingleParameterHypothesis::getrRSS( void ) const
{
    return this->m_rRSS;
}

void
SingleParameterHypothesis::setrRSS( double rrss )
{
    this->m_rRSS = rrss;
}


double
SingleParameterHypothesis::getAR2( void ) const
{
    return this->m_AR2;
}

void
SingleParameterHypothesis::setAR2( double ar2 )
{
    this->m_AR2 = ar2;
}

double
SingleParameterHypothesis::getSMAPE( void ) const
{
    return this->m_SMAPE;
}

void
SingleParameterHypothesis::setSMAPE( double smape )
{
    this->m_SMAPE = smape;
}

void
SingleParameterHypothesis::computeCost( const std::vector<DataPoint>& points )
{
    this->m_RSS  = 0;
    this->m_rRSS = 0;
    double smape = 0;
    for ( int i = 0; i < points.size(); i++ )
    {
        double predicted  = this->m_function->evaluate( points[ i ].getParameterList() );
        double actual     = points[ i ].getValue();
        double difference = predicted - actual;
        this->m_RSS += difference * difference;
        double relativeDifference = difference / points[ i ].getValue();
        this->m_rRSS += relativeDifference * relativeDifference;
        double abssum = fabs( actual ) + fabs( predicted );
        if ( abssum != 0.0 )
        {
            // This `if` condition prevents a division by zero, but it is correct: if sum is 0, both `actual` and `predicted`
            // must have been 0, and in that case the error at this point is 0, so we don't need to add anything.
            smape += fabs( difference ) / abssum * 2;
        }
    }
    this->m_SMAPE = smape / points.size() * 100;
}

void
SingleParameterHypothesis::computeAdjustedRSquared( double TSS, const std::vector<DataPoint>& points )
{
    this->m_AR2 = 0.0;
    double adjR = 1.0 - ( this->m_RSS / TSS );

    double degrees_freedom = points.size() - this->m_function->getCompoundTerms().size() - 1;

    this->m_AR2 = ( 1.0 - ( 1.0 - adjR ) * ( points.size() - 1.0 ) / degrees_freedom );

    return;
}

void
SingleParameterHypothesis::computeCostLeaveOneOutCrossValidation( const std::vector<DataPoint>& points, DataPoint missing )
{
    double predicted  = this->m_function->evaluate( missing.getParameterList() );
    double actual     = missing.getValue();
    double difference = predicted - actual;
    this->m_RSS += difference * difference;
    double relativeDifference = difference / missing.getValue();
    this->m_rRSS += relativeDifference * relativeDifference;
    double abssum = fabs( actual ) + fabs( predicted );
    if ( abssum != 0.0 )
    {
        this->m_SMAPE += ( fabs( difference ) / abssum * 2 ) / points.size() * 100;
    }
}


void
SingleParameterHypothesis::estimateParameters( const std::vector<DataPoint>& points )
{
    assert( m_function != NULL );

    int                 hypothesisTotalTerms = this->m_function->getCompoundTerms().size() + 1;
    int                 experimentCount      = points.size();
    double*             M                    = new double[ hypothesisTotalTerms * hypothesisTotalTerms ];
    double*             f                    = new double[ hypothesisTotalTerms ];
    double*             rhs                  = new double[ experimentCount ];
    double*             regularization       = new double[ hypothesisTotalTerms ];
    std::vector<double> AT;
    AT.resize( hypothesisTotalTerms * experimentCount );

    for ( int i = 0; i < AT.size(); i++ )
    {
        AT[ i ] = 0;
    }
    for ( int i = 0; i < hypothesisTotalTerms; i++ )
    {
        int index = i * experimentCount;
        for ( int rhs_index = 0; rhs_index < experimentCount; rhs_index++ )
        {
            AT[ index ] = 1;

            double eval;
            if ( i > 0 )
            {
                CompoundTerm& compoundTerm = ( this->m_function->getCompoundTerms() )[ i - 1 ];
                compoundTerm.setCoefficient( 1 ); // reset coefficient
                eval = compoundTerm.evaluate( points[ rhs_index ].getParameterList() );
            }
            else
            {
                eval = 1;
            }
            AT[ index ] *= eval;
            index++;
            rhs[ rhs_index ] = points[ rhs_index ].getValue();
        }
    }

    for ( int i = 0; i < hypothesisTotalTerms; i++ )
    {
        double reg = 1;
        int    tmp = floor( log10( AT[ experimentCount * i + experimentCount - 1 ] ) );
        if ( tmp > 0 )
        {
            reg = pow( 10.0f, tmp );
        }
        regularization[ i ] = reg;
        for ( int j = 0; j < experimentCount; j++ )
        {
            int index = experimentCount * i + j;
            AT[ index ] *= reg;
        }
    }

    for ( int i = 0; i < hypothesisTotalTerms; i++ )
    {
        int row = i * experimentCount;
        for ( int j = 0; j < hypothesisTotalTerms; j++ )
        {
            int t_row = experimentCount * j;
            int index = i * hypothesisTotalTerms + j;
            M[ index ] = 0;
            for ( int k = 0; k < experimentCount; k++ )
            {
                M[ index ] += AT[ row + k ] * AT[ t_row + k ];
            }
        }
    }
    for ( int i = 0; i < hypothesisTotalTerms; i++ )
    {
        int row = i * experimentCount;
        f[ i ] = 0;
        for ( int k = 0; k < experimentCount; k++ )
        {
            f[ i ] += AT[ row + k ] * rhs[ k ];
        }
    }

    matrixInverse( M, hypothesisTotalTerms );

    for ( int i = 0; i < hypothesisTotalTerms; i++ )
    {
        int irow = i * hypothesisTotalTerms;
        if ( i == 0 )
        {
            this->m_function->setConstantCoefficient( 0 );
        }
        else
        {
            ( this->m_function->getCompoundTerms() )[ i - 1 ].setCoefficient( 0 );
        }

        for ( int j = 0; j < hypothesisTotalTerms; j++ )
        {
            if ( i == 0 )
            {
                this->m_function->setConstantCoefficient( this->m_function->getConstantCoefficient() +  ( M[ irow + j ] * f[ j ] ) );
            }
            else
            {
                ( this->m_function->getCompoundTerms() )[ i - 1 ].setCoefficient( ( this->m_function->getCompoundTerms() )[ i - 1 ].getCoefficient() + ( M[ irow + j ] * f[ j ] ) );
            }
        }
        if ( i == 0 )
        {
            this->m_function->setConstantCoefficient( this->m_function->getConstantCoefficient() * regularization[ i ] );
        }
        else
        {
            ( this->m_function->getCompoundTerms() )[ i - 1 ].setCoefficient( ( this->m_function->getCompoundTerms() )[ i - 1 ].getCoefficient() * regularization[ i ] );
        }
    }

    delete[] M;
    delete[] f;
    delete[] rhs;
    delete[] regularization;
    AT.clear();
    return;
}

bool
SingleParameterHypothesis::isValid( void )
{
    double cost = m_RSS;
    return !( cost != cost || fabs( cost ) == std::numeric_limits<double>::infinity() );
}

double
SingleParameterHypothesis::calculateMaximalTermContribution( int termIndex, const std::vector<DataPoint>& points ) const
{
    std::vector<EXTRAP::CompoundTerm> ct = m_function->getCompoundTerms();
    assert( termIndex < ct.size() );
    double max = 0;
    for ( int i = 0; i < points.size(); i++ )
    {
        double contribution = fabs( ct[ termIndex ].evaluate( points[ i ].getParameterList() ) / points[ i ].getValue() );
        if ( contribution > max )
        {
            max = contribution;
        }
    }
    return max;
}

void
SingleParameterHypothesis::cleanConstantCoefficient( double epsilon, const std::vector<DataPoint>& points )
{
    // This function is used to correct numerical imprecision in the matrix caculations,
    // when the constant coefficient should be zero but is instead very small.
    // We take into account the minimum data value to make sure that we don't "nullify"
    // actually relevant numbers.
    std::vector<DataPoint>::const_iterator it      = points.begin();
    double                                 minimum = it->getValue();
    ++it;
    for (; it != points.end(); ++it )
    {
        if ( it->getValue() < minimum )
        {
            minimum = it->getValue();
        }
    }

    if ( fabs( this->m_function->getConstantCoefficient() / minimum ) < epsilon )
    {
        this->m_function->setConstantCoefficient( 0 );
    }
}

void
SingleParameterHypothesis::keepFunctionAlive( void )
{
    assert( this->m_function_refcount != NULL );
    assert( this->m_function_refcount->func == this->m_function );
    this->m_function_refcount->count++;

    //DebugStream << "Incremented external function refcount (" << this->m_function << ") at " << this->m_function_refcount << " to " << this->m_function_refcount->count << "\n";
}

void
SingleParameterHypothesis::freeFunction( void )
{
    assert( this->m_function_refcount != NULL );
    assert( this->m_function_refcount->func == this->m_function );
    ( this->m_function_refcount->count )--;
    //DebugStream << "Decremented external function refcount (" << this->m_function << ") at " << this->m_function_refcount << " to " << this->m_function_refcount->count << "\n";
    if ( this->m_function_refcount->count == 0 )
    {
        DebugStream << "### Deleting function (" << this->m_function << ") with refcount at " << this->m_function_refcount << "\n";
        delete this->m_function;
        delete this->m_function_refcount;
        this->m_function          = NULL;
        this->m_function_refcount = NULL;
    }
}
};
