/*
 * This file is part of the Extra-P software (http://www.scalasca.org/software/extra-p)
 *
 * Copyright (c) 2017,
 * Technische Universitaet Darmstadt, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 *
 */
#include <config.h>
#include <EXTRAP_SingleParameterModelGenerator.hpp>
#include <EXTRAP_SingleParameterSimpleFunctionModeler.hpp>
#include <EXTRAP_SingleParameterHypothesis.hpp>
#include <EXTRAP_SingleParameterFunction.hpp>
#include <EXTRAP_Interface.hpp>
#include <EXTRAP_Utilities.hpp>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cassert>

namespace EXTRAP
{
SingleParameterModelGenerator::SingleParameterModelGenerator()
{
    m_options = GENERATE_MODEL_MEAN;
}

SingleParameterModelGenerator::~SingleParameterModelGenerator()
{
}

Model*
SingleParameterModelGenerator::createModel( const ExperimentPointList& modeledPointList,
                                            const Function*            expectationFunction )
{
    FunctionPair* confidence_interval = new FunctionPair();
    FunctionPair* error_cone          = new FunctionPair();
    FunctionPair* noise_error         = new FunctionPair();

    confidence_interval->upper = new Function();
    confidence_interval->lower = new Function();
    error_cone->upper          = new Function();
    error_cone->lower          = new Function();
    noise_error->upper         = new Function();
    noise_error->lower         = new Function();

    // Build list of data points from experiment points
    std::vector<DataPoint> dataPoints;
    dataPoints.reserve( modeledPointList.size() );
    for ( std::vector<ExperimentPoint*>::const_iterator it = modeledPointList.begin(); it != modeledPointList.end(); ++it )
    {
        ExperimentPoint* p = *it;
        if ( m_options == GENERATE_MODEL_MEAN )
        {
            dataPoints.push_back( DataPoint( &( p->getCoordinate() ), p->getSampleCount(), p->getMean(), p->getMeanCI() ) );
        }
        else if ( m_options == GENERATE_MODEL_MEDIAN )
        {
            dataPoints.push_back( DataPoint( &( p->getCoordinate() ), p->getSampleCount(), p->getMedian(), p->getMedianCI() ) );
        }
        else
        {
            assert( false );
        }
    }
    ModelCommentList          comments       = ModelCommentList();
    SingleParameterHypothesis bestHypothesis = getFunctionModeler().createModel( dataPoints, comments, expectationFunction );

    // The resulting model will take ownership of the function
    return new Model( ( Function* )bestHypothesis.getFunction(),
                      this,
                      confidence_interval,
                      error_cone,
                      noise_error,
                      comments,
                      bestHypothesis.getRSS(),
                      bestHypothesis.getAR2(),
                      bestHypothesis.getSMAPE() );
}

void
SingleParameterModelGenerator::setCrossvalidationMethod( Crossvalidation cvMethod )
{
    this->getFunctionModeler().setCrossvalidationMethod( cvMethod );
}

Crossvalidation
SingleParameterModelGenerator::getCrossvalidationMethod() const
{
    return this->getFunctionModeler().getCrossvalidationMethod();
}

void
SingleParameterModelGenerator::setEpsilon( Value eps )
{
    this->getFunctionModeler().setEpsilon( eps );
}

Value
SingleParameterModelGenerator::getEpsilon( void ) const
{
    return this->getFunctionModeler().getEpsilon();
}

void
SingleParameterModelGenerator::setModelOptions( GenerateModelOptions options )
{
    this->m_options = options;
}

GenerateModelOptions
SingleParameterModelGenerator::getModelOptions( void ) const
{
    return this->m_options;
}

bool
SingleParameterModelGenerator::serialize( IoHelper* ioHelper ) const
{
    SAFE_RETURN( ioHelper->writeString(  this->getUserName() ) );
    Crossvalidation cv = this->getCrossvalidationMethod();
    std::string     cvMethod;
    switch ( cv )
    {
        case CROSSVALIDATION_NONE:
            cvMethod = "CROSSVALIDATION_NONE";
            break;
        case CROSSVALIDATION_LEAVE_ONE_OUT:
            cvMethod = "CROSSVALIDATION_LEAVE_ONE_OUT";
            break;
        case CROSSVALIDATION_LEAVE_P_OUT:
            cvMethod = "CROSSVALIDATION_LEAVE_P_OUT";
            break;
        case CROSSVALIDATION_K_FOLD:
            cvMethod = "CROSSVALIDATION_K_FOLD";
            break;
        case CROSSVALIDATION_TWO_FOLD:
            cvMethod = "CROSSVALIDATION_TWO_FOLD";
            break;
        default:
            ErrorStream << "Invalid Crossvalidation Method found in ModelGenerator. Exiting..." << std::endl;
            return false;
    }
    SAFE_RETURN( ioHelper->writeString(  cvMethod ) );
    SAFE_RETURN( ioHelper->writeValue(  this->getEpsilon() ) );
    //Write GenerateModelOptions
    GenerateModelOptions options = this->getModelOptions();
    std::string          optionsString;
    switch ( options )
    {
        case GENERATE_MODEL_MEAN:
            optionsString = "GENERATE_MODEL_MEAN";
            break;
        case GENERATE_MODEL_MEDIAN:
            optionsString = "GENERATE_MODEL_MEDIAN";
            break;
        default:
            ErrorStream << "Invalid GenerateModelOptions found in ModelGenerator. Exiting..." << std::endl;
            return false;
    }
    SAFE_RETURN( ioHelper->writeString(  optionsString ) );
    return true;
}

void
SingleParameterModelGenerator::deserialize(  IoHelper* ioHelper )
{
    std::string userName = ioHelper->readString();
    this->setUserName( userName );
    std::string cvMethod = ioHelper->readString();
    if ( cvMethod.compare( "CROSSVALIDATION_NONE" ) == 0 )
    {
        this->setCrossvalidationMethod( CROSSVALIDATION_NONE );
    }
    else if ( cvMethod.compare( "CROSSVALIDATION_LEAVE_ONE_OUT" ) == 0 )
    {
        this->setCrossvalidationMethod( CROSSVALIDATION_LEAVE_ONE_OUT );
    }
    else if ( cvMethod.compare( "CROSSVALIDATION_LEAVE_P_OUT" ) == 0 )
    {
        this->setCrossvalidationMethod( CROSSVALIDATION_LEAVE_P_OUT );
    }
    else if ( cvMethod.compare( "CROSSVALIDATION_K_FOLD" ) == 0 )
    {
        this->setCrossvalidationMethod( CROSSVALIDATION_K_FOLD );
    }
    else if ( cvMethod.compare( "CROSSVALIDATION_TWO_FOLD" ) == 0 )
    {
        this->setCrossvalidationMethod( CROSSVALIDATION_TWO_FOLD );
    }
    else
    {
        ErrorStream << "Invalid Crossvalidation Method found in File. Defaulting to No crossvalidation." << std::endl;
        this->setCrossvalidationMethod( CROSSVALIDATION_NONE );
    }
    Value eps = ioHelper->readValue();
    this->setEpsilon( eps );
    std::string modelOptions = ioHelper->readString();
    if ( modelOptions.compare( "GENERATE_MODEL_MEAN" ) == 0 )
    {
        this->setModelOptions( GENERATE_MODEL_MEAN );
    }
    else if ( modelOptions.compare( "GENERATE_MODEL_MEDIAN" ) == 0 )
    {
        this->setModelOptions( GENERATE_MODEL_MEDIAN );
    }
    else
    {
        ErrorStream << "Invalid ModelOptions found in File. Using Mean" << std::endl;
        this->setModelOptions( GENERATE_MODEL_MEAN );
    }
}
}; // Close namespace
